/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2015 ResiliNets, ITTC, University of Kansas
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Truc Anh N. Nguyen <annguyen@ittc.ku.edu>
 *
 * James P.G. Sterbenz <jpgs@ittc.ku.edu>, director
 * ResiliNets Research Group  http://wiki.ittc.ku.edu/resilinets
 * Information and Telecommunication Technology Center (ITTC)
 * and Department of Electrical Engineering and Computer Science
 * The University of Kansas Lawrence, KS USA.
 */

#include "tcp-general-test.h"
#include "ns3/node.h"
#include "ns3/log.h"
#include "ns3/private/tcp-option-sack.h"

namespace ns3 {

NS_LOG_COMPONENT_DEFINE ("SackTestSuite");

class SackTestCase : public TcpGeneralTest
{
public:
  enum Configuration
  {
    DISABLED,
    ENABLED_RECEIVER,
    ENABLED_SENDER,
    ENABLED
  };

  SackTestCase (SackTestCase::Configuration conf);
protected:
  virtual Ptr<TcpSocketMsgBase> CreateReceiverSocket (Ptr<Node> node);
  virtual Ptr<TcpSocketMsgBase> CreateSenderSocket (Ptr<Node> node);

  virtual void Tx (const Ptr<const Packet> p, const TcpHeader& h, SocketWho who);

  Configuration m_configuration;
};

SackTestCase::SackTestCase (SackTestCase::Configuration conf)
  : TcpGeneralTest ("Testing the TCP SACK option")
{
  m_configuration = conf;
}

Ptr<TcpSocketMsgBase>
SackTestCase::CreateReceiverSocket (Ptr<Node> node)
{
  Ptr<TcpSocketMsgBase> socket = TcpGeneralTest::CreateReceiverSocket (node);

  switch (m_configuration)
    {
    case DISABLED:
      socket->SetAttribute ("SACK", BooleanValue (false));
      break;

    case ENABLED_RECEIVER:
      socket->SetAttribute ("SACK", BooleanValue (true));
      break;

    case ENABLED_SENDER:
      socket->SetAttribute ("SACK", BooleanValue (false));
      break;

    case ENABLED:
      socket->SetAttribute ("SACK", BooleanValue (true));
    }

  return socket;
}

Ptr<TcpSocketMsgBase>
SackTestCase::CreateSenderSocket (Ptr<Node> node)
{
  Ptr<TcpSocketMsgBase> socket = TcpGeneralTest::CreateSenderSocket (node);

  switch (m_configuration)
    {
    case DISABLED:
      socket->SetAttribute ("SACK", BooleanValue (false));
      break;

    case ENABLED_RECEIVER:
      socket->SetAttribute ("SACK", BooleanValue (false));
      break;

    case ENABLED_SENDER:
      socket->SetAttribute ("SACK", BooleanValue (true));
      break;

    case ENABLED:
      socket->SetAttribute ("SACK", BooleanValue (true));
      break;
    }

  return socket;
}

void
SackTestCase::Tx (const Ptr<const Packet> p, const TcpHeader &h, SocketWho who)
{
  if (h.GetFlags () & TcpHeader::SYN)
    {
      if (m_configuration == DISABLED)
        {
          NS_TEST_ASSERT_MSG_EQ (h.HasOption (TcpOption::SACKPERMITTED), false,
                                 "SACK disabled but option enabled");
        }
      else if (m_configuration == ENABLED)
        {
          NS_TEST_ASSERT_MSG_EQ (h.HasOption (TcpOption::SACKPERMITTED), true,
                                 "SACK enabled but option disabled");
        }

      if (who == SENDER)
        {
          if (m_configuration == ENABLED_RECEIVER)
            {
              NS_TEST_ASSERT_MSG_EQ (h.HasOption (TcpOption::SACKPERMITTED), false,
                                     "SACK disabled but option enabled");
            }
          else if (m_configuration == ENABLED_SENDER)
            {
              NS_TEST_ASSERT_MSG_EQ (h.HasOption (TcpOption::SACKPERMITTED), true,
                                     "SACK enabled but option disabled");
            }
        }
      else if (who == RECEIVER)
        {
          if (m_configuration == ENABLED_RECEIVER)
            {
              NS_TEST_ASSERT_MSG_EQ (h.HasOption (TcpOption::SACKPERMITTED), false,
                                     "sender has not SACK, but receiver sent anyway");
            }
          else if (m_configuration == ENABLED_SENDER)
            {
              NS_TEST_ASSERT_MSG_EQ (h.HasOption (TcpOption::SACKPERMITTED), false,
                                     "receiver has not SACK enabled but sent anyway");
            }
        }
    }
  else
    {
      NS_TEST_ASSERT_MSG_EQ (h.HasOption (TcpOption::SACKPERMITTED), false,
                             "Sack-Permitted appended to non-SYN");
    }
}

static class TcpSackTestSuite : public TestSuite
{
public:
  TcpSackTestSuite ()
    : TestSuite ("tcp-option-sack", UNIT)
  {
    AddTestCase (new SackTestCase (SackTestCase::DISABLED), TestCase::QUICK);
    AddTestCase (new SackTestCase (SackTestCase::ENABLED_RECEIVER), TestCase::QUICK);
    AddTestCase (new SackTestCase (SackTestCase::ENABLED_SENDER), TestCase::QUICK);
    AddTestCase (new SackTestCase (SackTestCase::ENABLED), TestCase::QUICK);

  }

} g_tcpSackTestSuite;



} // namespace ns3
